@@ -1,4 +1,5 @@
 from ._models.sample import Sample
+# noinspection PyProtectedMember
 from ._models.transforms._matrix import Matrix
 from ._models import transforms
 from ._models import gates
@@ -7,7 +8,9 @@ from ._models.session import Session
 from ._models.dimension import Dimension, RatioDimension, QuadrantDivider
 from ._models.vertex import Vertex
 from ._utils.xml_utils import parse_gating_xml, export_gatingml
-from ._utils.plot_utils import plot_channel
+from ._utils.wsp_utils import parse_wsp
+from ._utils.plot_utils import plot_channel, calculate_extent
+from ._utils.sample_utils import load_samples, calculate_compensation_from_beads
 
 __all__ = [
     'Sample',
@@ -22,5 +25,9 @@ __all__ = [
     'transforms',
     'parse_gating_xml',
     'export_gatingml',
-    'plot_channel'
+    'parse_wsp',
+    'plot_channel',
+    'calculate_extent',
+    'load_samples',
+    'calculate_compensation_from_beads'
 ]
